local cube_model = game.ref("mdl", "cube")

local terrain_model = game.ref("mdl", "terrain")
local terrain_phys = game.ref("physics", "terrain")

local skybox = game.ref("tex", "sky1.cub.png")

local bgm_source = game.ref("stream", "background.ogg")
local bgm = game.mix.sound(bgm_source)

-- Map buttons to controls
game.controls["forward"] = "W"
game.controls["back"] = "S"
game.controls["left"] = "A"
game.controls["right"] = "D"
game.controls["jump"] = " "

return {
	load = function()
		
		game.setphys(terrain_phys)
		game.skybox(skybox)
		game.firstperson(true)
		
		bgm.volume = 0.3
		bgm.loop = true
		game.mix.play(bgm)
		
		local PLAYER_ENTITY = game.addentity{
			render = {},
			physics = {
				dynamics = "dynamic",
				capsule = {
					radius = 0.24,
					length = 1.5
				},
				pos = {0, 1, 0},
				speed = 5,
				trigger = 0
			},
		}
		
		function game.render(dt, alpha)
			local t = os.time()
			
			game.batch(cube_model, {math.sin(t) * 10, 3, math.cos(t) * 10})
			
			game.batch(terrain_model, {0, 0, 0})
		end
		
		function game.join(player)
			player:assign(PLAYER_ENTITY)
		end
		
		function game.update()
			-- Prevent player character from falling down abyss
			local c = game.getcomponent(PLAYER_ENTITY, "physics")
			if c.pos[2] < -5 then
				c.pos = {0, 1, 0}
			end
		end
		
		game.lights = {
			{"omni", position = {0, 20, 0}, color = {0.6, 0.6, 1, 0}, radius = 30}
		}
		
	end
}
