#version 430

#define MAX_POINTS (1500000)

#define BIAS 0.000001

layout(r32ui, binding = 0) uniform uimage2D u_multimap;
layout(r32ui, binding = 1) uniform uimage2D u_pixelsinshadow;

layout(std430, binding = 0) buffer Elements {
	uint ElNextFree;
	
	uint ElNexts[MAX_POINTS];
	vec3 ElPoints[MAX_POINTS];
	uint ElFBCoords[MAX_POINTS];
};

uniform mat4 u_lightvp;

uniform int u_lightnum;

varying vec4 v_gvertex0;
varying vec4 v_gvertex1;
varying vec4 v_gvertex2;

uniform float u_time;

out vec4 fragcolor;

// Get parametric coordinates of a triangle for point p
vec2 get_parametrics(vec2 p, vec2 v1, vec2 v2, vec2 v3) {
	return vec2(
		(p.x * (v3.y - v1.y) + p.y * (v1.x - v3.x) - v1.x * v3.y + v1.y * v3.x)
			/ +(v1.x * (v2.y - v3.y) + v1.y * (v3.x - v2.x) + v2.x * v3.y - v2.y * v3.x),
		
		(p.x * (v2.y - v1.y) + p.y * (v1.x - v2.x) - v1.x * v2.y + v1.y * v2.x)
			/ -(v1.x * (v2.y - v3.y) + v1.y * (v3.x - v2.x) + v2.x * v3.y - v2.y * v3.x)
	);
}

void main() {
	uint idx = imageLoad(u_multimap, ivec2(gl_FragCoord.xy)).r;
	
	float minz = 10;
	
	while(idx != 0xFFFFFFFF) {
		vec3 point = ElPoints[idx];
		
		minz = min(minz, point.z);
		
		vec2 params = get_parametrics(point.xy, v_gvertex0.xy, v_gvertex1.xy, v_gvertex2.xy);
		
		if(params.x >= 0 && params.x <= 1 && params.y >= 0 && params.y <= 1 && params.x + params.y <= 1) {
			// Point is inside the triangle in 2D
			
			// Now we must determine which of the two is closer
			
			// Compute the Z of the triangle at the same point
			float z = v_gvertex0.z + params.x * (v_gvertex1.z - v_gvertex0.z) + params.y * (v_gvertex2.z - v_gvertex0.z);
			
			if(z + BIAS <= point.z) {
				// Point is behind the triangle
				
				// Decode the main view pixel coordinates
				uint fbcoordPacked = ElFBCoords[idx];
				ivec2 fbcoord = ivec2(fbcoordPacked >> 16, fbcoordPacked & uint(0xFFFF));
				
				// Mark main view pixel as shadowed
				imageAtomicOr(u_pixelsinshadow, fbcoord, 1 << u_lightnum);
			}
		}
		
		idx = ElNexts[idx];
	}
	
	if(minz != 10) {
		float g = 2.0 * 0.1 * 30.0 / (30.0 + 0.1 - minz * (30.0 - 0.1));
		
		fragcolor = vec4(vec3(0, g / 30.0, 0), 1);
	} else {
		float g = 2.0 * 0.1 * 30.0 / (30.0 + 0.1 - gl_FragCoord.z * (30.0 - 0.1));
		
		fragcolor = vec4(g / 30.0, 0, 0, 1);
	}
}
