#extension GL_ARB_shader_image_load_store : enable

#ifdef GLCORE
#define varying in
#define gl_FragColor fragcol
out vec4 fragcol;
in vec2 v_uv;
uniform mat4 u_projection;
#else
#define u_projection gl_ProjectionMatrix
#define v_uv gl_TexCoord[0].xy
#endif

uniform sampler2D u_dif;

#ifdef NORMAL
uniform sampler2D u_nrm;
#endif

#ifdef DISPLACEMENT
uniform sampler2D u_dsp;
#endif

#ifdef EMISSION
uniform sampler2D u_emt;
#endif

#ifdef ROUGHNESS
uniform sampler2D u_rgh;
#endif

uniform mat4 u_view;
uniform vec3 u_cam;

varying vec3 v_pos;
varying vec3 v_nrm;
varying vec3 v_tan;
varying vec4 v_col;

varying vec3 v_posworld;

#define k3_MAX_LIGHTS_PER_PASS 4
#define k3_LIGHT_TYPE(i) u_BaseLightSettings1[i].w
uniform vec4 u_BaseLightSettings1[k3_MAX_LIGHTS_PER_PASS];
uniform vec4 u_BaseLightSettings2[k3_MAX_LIGHTS_PER_PASS];
uniform vec4 u_BaseLightColors[k3_MAX_LIGHTS_PER_PASS];
uniform vec4 u_AmbientLight;

uniform float u_time;

//#include klighting.glsl

float distsq(vec3 a, vec3 b) {
	vec3 d = b - a;
	return dot(d, d);
}

#ifdef DISPLACEMENT

#ifndef DISP_HEIGHT
#define DISP_HEIGHT 0.025
#endif

#ifndef DISP_PARALLAX_MAX
#define DISP_PARALLAX_MAX 15.0
#endif

vec2 parallax(vec2 tc, vec3 dir) {
	float dh = mix(DISP_HEIGHT, 0.0, smoothstep(DISP_PARALLAX_MAX - 5.0, DISP_PARALLAX_MAX, length(v_pos)));
	if(dh <= 0.0005) return tc;
	float numLayers = mix(40.0, 80.0, max(0.0, dot(vec3(0.0, 0.0, 1.0), dir)));
	float layerDepth = 1.0 / numLayers;
	float currentLayerDepth = 0.0;
	vec2 P = dir.xy * dh;
	vec2 deltaTexCoords = P / numLayers;
	vec2 currentTexCoords = tc;
	float currentDepthMapValue = texture2D(u_dsp, currentTexCoords).r;
	for(int i = 0; i < 100; i++) {
		if(currentLayerDepth >= currentDepthMapValue) break;
		currentTexCoords -= deltaTexCoords;
		currentDepthMapValue = texture2D(u_dsp, currentTexCoords).r;
		currentLayerDepth += layerDepth;
	}
	vec2 prevTexCoords = currentTexCoords + deltaTexCoords;
	float afterDepth = currentDepthMapValue - currentLayerDepth;
	float beforeDepth = texture2D(u_dsp, prevTexCoords).r - currentLayerDepth + layerDepth;
	float weight = afterDepth / (afterDepth - beforeDepth);
	return prevTexCoords * weight + currentTexCoords * (1.0 - weight);
}
#endif

struct Lighting {
	vec3 diffuse;
	vec3 specular;
};

uniform sampler2D u_shadows0atlas;

#define k3_MAX_SHADOWVPS (k3_MAX_LIGHTS_PER_PASS + 2)

uniform mat4 u_shadows0vp[k3_MAX_SHADOWVPS];
uniform vec4 u_shadows0seg[k3_MAX_LIGHTS_PER_PASS];
bool testshadow(int light, int layer, vec3 transf, float bias) {
	vec4 seg = u_shadows0seg[light];
	
	float cellID = seg.x + layer;
	float cellSz = seg.y;
	
	float cellX = mod(cellID * cellSz, 1);
	float cellY = floor(cellID * cellSz) * cellSz;
	
	vec2 point = vec2(cellX + transf.x * cellSz, cellY + transf.y * cellSz);
	
	if(transf.z - bias > texture2D(u_shadows0atlas, point).r) {
		return false;
	}
	
	return true;
}

int lightVPCount(int light) {
	if(k3_LIGHT_TYPE(light) == 0) {
		return int(u_BaseLightSettings2[light].x);
	}
	return 1;
}

vec2 poissonDisk[64] = vec2[] (
	vec2(-0.6402387640985373, 0.01239059148005106),
	vec2(-0.7928810939713299, -0.27046990600667453),
	vec2(0.34745047759894776, 0.15976667514949217),
	vec2(0.450661918145129, 0.1704823534253883),
	vec2(0.3911811662877689, 0.06738522386791956),
	vec2(-0.17665568510544727, 0.19625234243627498),
	vec2(-0.39306028868313014, 0.17443163037792675),
	vec2(0.8693300677690223, -0.5046355892880614),
	vec2(0.731501765350844, -0.7289125777474403),
	vec2(0.8603280870748453, -0.618904523551793),
	vec2(0.30491902729600096, 0.2945595829709575),
	vec2(0.4694863881903675, 0.2941268152227128),
	vec2(0.7237569448324053, -0.8364462591744583),
	vec2(0.16451313040820081, 0.3866766719587804),
	vec2(0.38046045705899956, 0.4495130224721353),
	vec2(0.26867122324555215, 0.40986661116765744),
	vec2(0.5120540770222164, 0.4858275091737214),
	vec2(0.33379999718232334, 0.6096670024179336),
	vec2(0.5386103428511821, 0.3677635329294022),
	vec2(0.4646776478934602, 0.5784010915310125),
	vec2(0.6485909713678137, -0.9152340003404229),
	vec2(0.8242141019926763, 0.14220894869442158),
	vec2(0.9536720508356535, 0.10976090469967748),
	vec2(-0.6847424934844475, 0.17599643949962784),
	vec2(-0.5455425291949105, 0.19258044933397),
	vec2(-0.4009164846053752, -0.7573717773303037),
	vec2(0.6336179906319963, 0.15877911467497396),
	vec2(0.6536376163094109, 0.2800712945345989),
	vec2(0.8682083042372091, -0.9158797534098212),
	vec2(0.7504235539461312, -0.9966221811879137),
	vec2(0.8796470159897856, -0.7324050282677629),
	vec2(-0.571176606877617, -0.5975017885015055),
	vec2(-0.43970047543729907, -0.5079476465439433),
	vec2(0.695334052581182, 0.459687368786893),
	vec2(-0.6161401668433959, 0.3459670656434264),
	vec2(0.41690653873434513, 0.707867506448423),
	vec2(0.5144088974115453, 0.750784605442163),
	vec2(0.6370698323293555, 0.56399799689076),
	vec2(0.5597066682451106, 0.653108466443086),
	vec2(-0.3820312657872147, -0.882739416712766),
	vec2(-0.518877538505339, -0.7197105875153247),
	vec2(-0.5268706934754644, -0.868488088700591),
	vec2(-0.6952540559560036, -0.6991715296439802),
	vec2(-0.6500902131772277, -0.7905449684775162),
	vec2(-0.880723364011579, -0.4993708838196964),
	vec2(-0.6707578387262574, -0.5708271319896198),
	vec2(0.8375753186950639, 0.29960999423301415),
	vec2(-0.4180721012732118, 0.29797505709386307),
	vec2(0.10193809763584283, 0.2523714015975511),
	vec2(0.1862282352998137, 0.5110376928389648),
	vec2(-0.0438929895360628, 0.33195669656728866),
	vec2(-0.13334897225417575, 0.3859871863414337),
	vec2(-0.274586030844141, 0.36300014759327914),
	vec2(-0.051560336739855384, 0.18831000978208534),
	vec2(-0.5087937841363511, -0.9844420022501683),
	vec2(-0.8048506977534813, -0.8460965549278475),
	vec2(-0.6994511562235357, -0.9765234753285633),
	vec2(-0.780160570099029, -0.6435410942147783),
	vec2(-0.15979453172633906, 0.5404332904078759),
	vec2(0.05310156021218737, 0.431707311989884),
	vec2(-0.05614457022156283, 0.45310708947191225),
	vec2(-0.29056378317928544, 0.46499253589320544),
	vec2(0.971345815224288, -0.806210231255247),
	vec2(-0.8552446952697963, -0.7484411633213446)
);

float random(vec4 v) {
	return fract(sin(dot(v, vec4(12.9898, 78.233, 45.164, 94.673))) * 43758.5453);
}

float k3TestShadow(int vpStart, int light, float bias) {
	int layer = 0;
	
	vec3 pos = v_posworld;
	
	if(k3_LIGHT_TYPE(light) == 3) {
		// Determine face
		vec3 lpos = u_BaseLightSettings1[light].xyz;
		vec3 diff = v_posworld - lpos;
		if(diff.x >= abs(diff.y) && diff.x >= abs(diff.z)) {
			layer = 3;
			pos = vec3(diff.z, diff.y, -diff.x) + lpos;
		} else if(diff.x <= -abs(diff.y) && diff.x <= -abs(diff.z)) {
			layer = 2;
			pos = vec3(-diff.z, diff.y, diff.x) + lpos;
		} else if(diff.z >= abs(diff.x) && diff.z >= abs(diff.y)) {
			layer = 1;
			pos = vec3(-diff.x, diff.y, -diff.z) + lpos;
		} else if(diff.z <= -abs(diff.x) && diff.z <= -abs(diff.y)) {
			layer = 0;
		} else if(diff.y >= abs(diff.x) && diff.y >= abs(diff.z)) {
			layer = 5;
			pos = vec3(-diff.x, diff.z, diff.y) + lpos;
		} else if(diff.y <= -abs(diff.x) && diff.y <= -abs(diff.z)) {
			layer = 4;
			pos = vec3(-diff.x, diff.z, diff.y) + lpos;
		}
	}
	
	int availableVPs = lightVPCount(light);
	
	vec4 transf;
	int vp = 0;
	
	for(; vp < availableVPs; vp++) {
		transf = u_shadows0vp[vpStart + vp] * vec4(pos, 1.0);
		transf.xyz = transf.xyz / transf.w * 0.5 + 0.5;
		
		if(transf.z >= 0.0 && transf.z <= 1.0 && transf.x >= 0.0 && transf.y >= 0.0 && transf.x <= 1.0 && transf.y <= 1.0) {
			break;
		}
	}
	
	if(vp == availableVPs) {
		return 1.0;
	}
	
	float asdf = 0.0;
	
	for(int x = 0; x < 16; x++) {
		int index = int(mod(64.0 * random(vec4(floor(v_posworld.xyz * 250.0), x)), 64.0));
		
		vec3 coord = transf.xyz;
		coord.xy += poissonDisk[index] / 1024.0;
		
		asdf += float(testshadow(light, layer + vp, coord, bias));
	}
	
	return asdf / 16.0;
}

float computeShadowBias(float ndotl) {
	float c = clamp(ndotl, 0.0, 1.0);
	float s = sqrt(1.0 - c * c);
	return 0.001 + 0.00001 * s / (c + 0.0000001);
}

Lighting do_lights(vec3 pos, vec3 nrm, vec3 geom_nrm, float roughness) {
	Lighting ret;
	
	ret.diffuse = vec3(0);
	ret.specular = vec3(0);
	
	int nextVP = 0;
	
	for(int i = 0; i < k3_MAX_LIGHTS_PER_PASS; i++) {
		float d = 0;
		float s = 0;
		
		float strength = 1.0;
		
		vec3 ldir;
		
		if(k3_LIGHT_TYPE(i) == 2 || k3_LIGHT_TYPE(i) == 3) {
			vec4 lpos = u_view * vec4(u_BaseLightSettings1[i].xyz, 1.0);
			ldir = normalize(lpos.xyz - v_pos);
			float radius = u_BaseLightSettings2[i].w;
			
			float dist = distance(lpos.xyz, v_pos);
			float attenuation = pow(clamp(1.0 - (dist * dist) / (radius * radius), 0.0, 1.0), 2.0);
			
			strength *= attenuation;
		} else if(k3_LIGHT_TYPE(i) == 1) {
			vec4 lpos = u_view * vec4(u_BaseLightSettings1[i].xyz, 1.0);
			ldir = normalize(lpos.xyz - v_pos);
			
			float inclination = u_BaseLightSettings2[i].x;
			float azimuth = u_BaseLightSettings2[i].y;
			
			vec3 d = (u_view * vec4(sin(inclination) * cos(azimuth), cos(inclination), sin(inclination) * sin(azimuth), 0.0)).xyz;
			
			if(dot(-ldir, d) < u_BaseLightSettings2[i].z) {
				strength = 0;
			} else {
				float radius = u_BaseLightSettings2[i].w;
				
				float dist = distance(lpos.xyz, v_pos);
				float attenuation = pow(clamp(1.0 - (dist * dist) / (radius * radius), 0.0, 1.0), 2.0);
				
				strength *= attenuation;
			}
		} else if(k3_LIGHT_TYPE(i) == 0) {
			ldir = -normalize((u_view * vec4(u_BaseLightSettings1[i].xyz, 0.0)).xyz);
		}
		
		float ndotl = max(0.0, dot(nrm, ldir));
		d = ndotl * strength;
		if(ndotl > 0.0) {
			s = min(pow(max(dot(normalize(ldir - normalize(v_pos)), nrm), 0.0), (1 - roughness) * (1 - roughness) * 64) / 4.0, 1.5) * d;
		}
		
		float bias = computeShadowBias(dot(geom_nrm, ldir));
		
		float shadowFactor = k3TestShadow(nextVP, i, bias) * 0.9;
		d *= shadowFactor;
		s *= shadowFactor;
		
		ret.diffuse += u_BaseLightColors[i].rgb * d;
		ret.specular += u_BaseLightColors[i].rgb * s;
		
		nextVP += lightVPCount(i);
	}
	
	ret.diffuse += u_AmbientLight.rgb;
	
	return ret;
}

mat3 transp(mat3 a) {
	return mat3(vec3(a[0].x, a[1].x, a[2].x), vec3(a[0].y, a[1].y, a[2].y), vec3(a[0].z, a[1].z, a[2].z));
}
void main() {
	vec3 n = normalize(v_nrm);
#if defined(NORMAL)
	vec3 t = normalize(v_tan);
	vec3 b = cross(n, t);
	if(dot(cross(t, b), n) < 0) {
		t = -t;
		b = -b;
	}
	mat3 tbn = mat3(t, b, n);
#if defined(DISPLACEMENT)
	vec2 tc = parallax(v_uv, normalize(transp(tbn) * (vec3(1.0, -1.0, 1.0) * normalize(v_pos))));
#else
	vec2 tc = v_uv;
#endif
	vec3 nrm = tbn * normalize(vec3(1.0, -1.0, 1.0) * (texture2D(u_nrm, tc).xyz * 2.0 - 1.0));
#else
	vec3 nrm = normalize(v_nrm);
	vec2 tc = v_uv;
#endif
	vec3 diffuse = vec3(0.0);
	vec3 specular = vec3(0.0);
#if defined(ROUGHNESS)
	float roughness = texture2D(u_rgh, tc).r;
#else
#ifdef FIXED_ROUGHNESS
	float roughness = FIXED_ROUGHNESS;
#else
	float roughness = 0;
#endif
#endif
	
	Lighting lighting = do_lights(v_pos, nrm, n, roughness);
	
	vec4 texdif = texture2D(u_dif, tc);
	
	gl_FragColor =	vec4((1.0 - roughness) * texdif.rgb * lighting.diffuse
				+	roughness * lighting.specular
#ifdef EMISSION
				+	texture2D(u_emt, tc).rgb, 1.0);
#else
#ifdef COLOR_AS_ALPHA
				, v_col.r);
#else
				, texdif.a * v_col.a);
#endif
#endif
}
