#ifdef GLCORE
uniform mat4 u_projection;
in vec3 a_pos;
in vec3 a_normal;
in vec2 a_uv;
in vec4 a_color;
#define varying out
#define attribute in
#define u_mv (u_view * u_model)
#define u_mvinvtranspose inverse(transpose(u_mv))
#define ftransform() (u_projection * u_view * u_model * vec4(a_pos, 1))
out vec2 v_uv;
#else
#define u_projection gl_ProjectionMatrix
#define a_pos gl_Vertex.xyz
#define a_normal gl_Normal.xyz
#define a_uv gl_MultiTexCoord0.xy
#define a_color gl_Color
#define u_mv gl_ModelViewMatrix
#define u_mvinvtranspose gl_ModelViewMatrixInverseTranspose
#define v_uv gl_TexCoord[0].xy
#endif

uniform mat4 u_view;
uniform mat4 u_model;

#ifdef k3_GPU_SKINNING
uniform vec4 u_bonedata[48];
attribute vec4 a_boneids;
attribute vec4 a_boneweights;
#endif

attribute vec3 a_tangent;

varying vec3 v_pos;
varying vec3 v_posworld;
varying vec3 v_nrm;
varying vec3 v_tan;
varying vec4 v_col;

vec3 rotquat(vec3 v, vec4 q) {
	vec3 u = q.xyz;
	float s = q.w;
	return 2.0 * dot(u, v) * u + (s * s - dot(u, u)) * v + 2.0 * s * cross(u, v);
}

void main() {
	v_col = a_color;
#if defined(k3_GPU_SKINNING)
	vec3 pos;
	pos  = (rotquat(a_pos, u_bonedata[1 + 2 * int(a_boneids[0])]) + u_bonedata[0 + 2 * int(a_boneids[0])].xyz) * a_boneweights[0];
	pos += (rotquat(a_pos, u_bonedata[1 + 2 * int(a_boneids[1])]) + u_bonedata[0 + 2 * int(a_boneids[1])].xyz) * a_boneweights[1];
	pos += (rotquat(a_pos, u_bonedata[1 + 2 * int(a_boneids[2])]) + u_bonedata[0 + 2 * int(a_boneids[2])].xyz) * a_boneweights[2];
	pos += (rotquat(a_pos, u_bonedata[1 + 2 * int(a_boneids[3])]) + u_bonedata[0 + 2 * int(a_boneids[3])].xyz) * a_boneweights[3];
	v_posworld = (u_model * vec4(pos, 1.0)).xyz;
	v_pos = (u_mv * vec4(pos, 1.0)).xyz;
	mat4 mv = u_mvinvtranspose;
	v_nrm = normalize(mat3(mv[0].xyz, mv[1].xyz, mv[2].xyz) * a_normal.xyz);
	v_tan = normalize(mat3(mv[0].xyz, mv[1].xyz, mv[2].xyz) * a_tangent.xyz);
	gl_Position = u_projection * vec4(v_pos, 1.0);
#else
	v_posworld = (u_model * vec4(a_pos, 1.0)).xyz;
	v_pos = (u_mv * vec4(a_pos, 1.0)).xyz;
	mat4 mv = u_mvinvtranspose;
	v_nrm = normalize(mat3(mv[0].xyz, mv[1].xyz, mv[2].xyz) * a_normal.xyz);
	v_tan = normalize(mat3(mv[0].xyz, mv[1].xyz, mv[2].xyz) * a_tangent.xyz);
	gl_Position = ftransform();
#endif
	
	v_uv = a_uv;
}
